local map = vim.keymap.set
local Terminal = require("toggleterm.terminal").Terminal
local notify = require("notify")
local dsplit = require("util.open-docs-in-split")

local function term_factory(cfg)
	cfg["on_stderr"] = function(_, job, data, name)
		notify(name .. " encountered an error on job: " .. job .. "\nData: " .. data, "error")
	end
	cfg["on_stdout"] = function(_, job, data, name)
		notify(name .. " output for job" .. job .. "\nData: " .. data, "info")
	end
	return Terminal:new(cfg)
end

local function term_toggle(term)
	term:toggle()
end

local opts = { silent = true, noremap = true }
local notsilent = { silent = false, noremap = true }

local function ReloadConfig()
	dofile(vim.fn.stdpath("config") .. "/init.lua")
	dofile(vim.fn.stdpath("config") .. "/lua/core/keymaps.lua")
	dofile(vim.fn.stdpath("config") .. "/lua/core/autocmds.lua")
	dofile(vim.fn.stdpath("config") .. "/lua/core/options.lua")
	dofile(vim.fn.stdpath("config") .. "/lua/core/plugins.lua")
	dofile(vim.fn.stdpath("config") .. "/lua/plugins/ui/init.lua")
	dofile(vim.fn.stdpath("config") .. "/lua/plugins/ui/toggleterm.lua")
	dofile(vim.fn.stdpath("config") .. "/lua/plugins/ai/init.lua")
	dofile(vim.fn.stdpath("config") .. "/lua/plugins/lsp/init.lua")
	dofile(vim.fn.stdpath("config") .. "/lua/plugins/colorscheme/init.lua")
end

-- Leader key
vim.g.mapleader = " "
vim.g.maplocalleader = ","

-- Custom commands
vim.api.nvim_create_user_command("PS", ":PackerSync", {})
vim.api.nvim_create_user_command("Reload", ReloadConfig, {})
vim.api.nvim_create_user_command("Config", "edit ~/.config/nvim", {})
vim.api.nvim_create_user_command("Plugins", "edit ~/.config/nvim/lua/core/plugins.lua", {})
vim.api.nvim_create_user_command("Keymaps", "edit ~/.config/nvim/lua/core/keymaps.lua", {})

-- {{{ Basic mappings
map("n", "<C-u>", "<C-u>zz", opts)
map("n", "n", "nzzzv", opts)
map("n", "N", "Nzzzv", opts)
map("x", "<leader>p", '"_dP', opts) -- paste without yanking
map("v", "<", "<gv", opts) -- reselect after indent
map("v", ">", ">gv", opts)
map("v", "J", ":m '>+1<CR>gv=gv", opts) -- move lines
map("v", "K", ":m '<-2<CR>gv=gv", opts)
-- }}}

-- {{{ Buffer navigation
map("n", "<C-J>", ":bnext<CR>", opts)
map("n", "<C-K>", ":bprev<CR>", opts)
map("n", "<leader>bb", ":Telescope buffers<CR>", opts)
map("n", "<leader>bk", ":bdelete<CR>", opts)
map("n", "<leader>bn", ":bnext<CR>", opts)
map("n", "<leader>bp", ":bprev<CR>", opts)
-- }}}

--{{{ Terminal mappings

local programs_map = {
	gg = { cmd = "lazygit", display_name = "lazygit", direction = "tab", hidden = true },
	op = { cmd = "ipython", display_name = "ipython", direction = "vertical", hidden = true },
	oP = { cmd = "ipython", display_name = "ipython-full", direction = "tab", hidden = true },
	on = { cmd = "ncmpcpp", display_name = "ncmpcpp", direction = "tab", hidden = true },
	ob = { cmd = "btop", display_name = "btop", direction = "tab", hidden = true },
	od = { cmd = "lazydocker", display_name = "lazydocker", direction = "tab", hidden = true },
}

for key, value in pairs(programs_map) do
	map("n", "<leader>" .. key, function()
		term_toggle(term_factory(value))
	end, opts)
end

map("v", "op", "<C-\\><C-N>:ToggleTerm name=ipython")
map("v", "oP", "<C-\\><C-N>:ToggleTerm name=ipython-full")

map("n", "<C-T>", ":ToggleTerm name=toggleterm<CR>", opts)
map("n", "<leader>tt", ":ToggleTerm name=toggleterm<CR>", opts)
map("n", "<leader>tT", ":ToggleTerm name=toggleterm-full direction=tab<CR>", opts)
map("n", "<leader>ot", ":ToggleTerm name=toggleterm<CR>", opts)
map("n", "<leader>oT", ":ToggleTerm name=toggleterm-full direction=tab<CR>", opts)
map("n", "<leader>ts", ":TermSelect<CR>", opts)
map("n", "<leader>tv", ":ToggleTerm direction=vertical name=toggleterm-vert<CR>", opts)
map("n", "<leader>th", ":ToggleTerm direction=horizontal name=toggleterm-hori<CR>", opts)
map("n", "<leader>ov", ":ToggleTerm direction=vertical name=toggleterm-vert<CR>", opts)
map("n", "<leader>oh", ":ToggleTerm direction=horizontal name=toggleterm-hori<CR>", opts)
map("n", "<leader>tf", ":ToggleTerm name=toggleterm<CR>", opts)
map("n", "<leader>-", ":ToggleTerm direction='horizontal'<CR>", opts)
map("n", "<leader>|", ":ToggleTerm direction='vertical'<CR>", opts)

function _G.set_terminal_keymaps()
	local opts = { buffer = 0 }
	map("t", "<esc>", [[<C-\><C-n>]], opts)
	map("t", "<C-w>", [[<C-\><C-n><C-w>]], opts)
	-- map("t", "<space>", "<space>", { buffer = 0, silent = true }) -- fix space in terminal
end

-- if you only want these mappings for toggle term use term://*toggleterm#* instead
vim.cmd("autocmd! TermOpen term://* lua set_terminal_keymaps()")
--}}}

--{{{ LSP mappings
map("n", "gA", vim.lsp.buf.code_action, opts)
map("n", "gd", ":Telescope lsp_definitions<CR>", opts)
map("n", "gDc", ":Telescope lsp_implementations<CR>", opts)
map("n", "gDf", ":Telescope lsp_definitions<CR>", opts)
map("n", "gF", ":edit <cfile><CR>", opts)
map("n", "gT", ":Telescope lsp_type_definitions<CR>", opts)
map("n", "gb", ":Gitsigns blame_line<CR>", opts)
map("n", "gi", ":Telescope lsp_implementations<CR>", opts)
map("n", "gj", ":Telescope jumplist<CR>", opts)
map("n", "gl", vim.lsp.codelens.run, opts)
map("n", "gr", ":Telescope lsp_references<CR>", opts)
map("n", "gs", vim.lsp.buf.signature_help, opts)
map("n", "K", vim.lsp.buf.hover, opts)
map("n", "<leader>kv", dsplit, opts)
map("n", "<leader>cd", ":Telescope lsp_definitions<CR>")
map("n", "<leader>ca", vim.lsp.buf.code_action)
map("n", "<leader>cci", ":Telescope lsp_incoming_calls<CR>")
map("n", "<leader>cco", ":Telescope lsp_outgoing_calls<CR>")
map("n", "<leader>ch", ":lua vim.lsp.buf.signature_help()<CR>")
map("n", "<leader>ci", ":Telescope lsp_implementations<CR>")
map("n", "<leader>cr", ":Telescope lsp_references<CR>")
map("n", "<leader>cR", ":lua vim.lsp.buf.rename()<CR>")
map("n", "<leader>cs", ":Telescope lsp_document_symbols<CR>")
map("n", "<leader>ct", ":Telescope lsp_type_definitions<CR>")
map("n", "<leader>cw", ":Telescope lsp_dynamic_workspace_symbols<CR>")
map("n", "<leader>cd", ":Telescope diagnostics theme=dropdown layout_config={width=0.8}<CR>")
map("n", "<leader>cDs", ":Telescope diagnostics theme=dropdown layout_config={width=0.8}<CR>")
map("n", "<leader>cDn", ":lua vim.diagnostic.goto_next()<CR>")
map("n", "<leader>cDp", ":lua vim.diagnostic.goto_prev()<CR<CR>")
map("n", "<leader>cl", ":lua vim.diagnostic.setloclist()<CR>")
map("n", "<leader>cH", open_doc_in_vsplit, opts)
--}}}

--{{{ Code Companion and Copilot
map("n", "<leader>cp", ":vert Copilot panel<CR>", opts)
map("n", "<leader>oc", ":CodeCompanionChat Toggle<CR>")
map("n", "<leader>Cc", ":CodeCompanionChat Toggle<CR>", opts)
map("n", "<leader>Ci", ":CodeCompanion ", opts)
map("n", "<leader>CT", ":CodeCompanionChat Toggle<CR>", opts)
map("n", "<leader>Ca", ":CodeCompanionActions<CR>", opts)
map("v", "<leader>Cc", ":CodeCompanionChat Add<CR>", opts)
map("v", "<leader>Ci", ":CodeCompanion /buffer ", notsilent)
map("v", "<leader>Ce", ":CodeCompanion /explain<CR>", notsilent)
map("v", "<leader>Cf", ":CodeCompanion /fix<CR>", notsilent)
map("v", "<leader>Cl", ":CodeCompanion /lsp<CR>", notsilent)
map("v", "<leader>Ct", ":CodeCompanion /tests<CR>", notsilent)
--}}}

--{{{ Telescope mappings
-- {{{ Telescope Finders
map("n", "//", ":Telescope current_buffer_fuzzy_find previewer=false<CR>", opts)
map("n", "??", ":Telescope lsp_document_symbols theme=dropdown layout_config={width=0.5}<CR>", opts)
map(
	"n",
	"<leader>fc",
	':Telescope color_names theme=dropdown layout_config={width=0.45,height=25,prompt_position="bottom"} layout_strategy=vertical<CR>',
	opts
)
map(
	"n",
	"<leader>Tc",
	':Telescope color_names theme=dropdown layout_config={width=0.45,height=25,prompt_position="bottom"} layout_strategy=vertical<CR>',
	opts
)
map("n", "<leader>Tn", ":Telescope notify theme=dropdown layout_config={width=0.75}<CR>", opts)

map(
	"n",
	"<leader>ff",
	":Telescope find_files find_command=rg,--ignore,--follow,--hidden,--files prompt_prefix=🔍<CR>",
	opts
)
map("n", "<leader>fg", ":Telescope live_grep<CR>", opts)
map(
	"n",
	"<leader>Tg",
	':Telescope glyph theme=dropdown layout_config={width=0.45,height=35,prompt_position="bottom"} layout_strategy=vertical<CR>',
	opts
)
map(
	"n",
	"<leader>fG",
	':Telescope glyph theme=dropdown layout_config={width=0.45,height=35,prompt_position="bottom"} layout_strategy=vertical<CR>',
	opts
)
map("n", "<leader>fb", ":Telescope file_browser<CR>", opts)
map("n", "<leader>fr", ":Telescope oldfiles theme=dropdown layout_config={width=0.5}<CR>", opts)
-- }}}

--{{{ Telescope Help
map("n", "<leader>hc", ":Telescope commands<CR>")
map("n", "<leader>hv", ":Telescope vim_options<CR>")
map("n", "<leader>hk", ":Telescope keymaps<CR>")
map("n", "<leader>hs", ":Telescope spell_suggest<CR>")
map("n", "<leader>hm", ":Telescope man_pages theme=dropdown layout_config={width=0.75}<CR>")
--}}}

--{{{ Telescope Search
map(
	"n",
	"<leader>sf",
	":Telescope find_files find_command=rg,--ignore,--follow,--hidden,--files prompt_prefix=🔍<CR>",
	opts
)
map("n", "<leader>sF", ":Telescope fidget<CR>")
map("n", "<leader>sg", ":Telescope live_grep<CR>")
map("n", "<leader>sh", ":Telescope command_history<CR>")
map("n", "<leader>sm", ":Telescope man_pages<CR>")
map("n", "<leader>s/", ":Telescope search_history<CR>")
--}}}

--{{{ Git mappings
map("n", "<leader>gc", ":Telescope git_commits<CR>", opts)
map("n", "<leader>gf", ":Telescope git_files<CR>", opts)
--}}}

map("n", "<leader>Tr", ":Telescope reloader<CR>", opts)

map("n", "Q", ":Telescope cmdline<CR>", opts)
map("n", "<leader><leader>", ":Telescope cmdline<CR>", opts)
--}}}

--{{{ File explorer and tools
map("n", "<leader>n", ":NvimTreeToggle<CR>", opts)
map("n", "<leader>D", ":Dotenv .env<CR>", opts)
--}}}

--{{{ Misc utilities
map("n", "<leader>x", "<cmd>!chmod +x %<CR>", opts)
map("n", "<leader>y", '"+', opts)
map("v", "<leader>y", '"+', opts)
map("n", "<leader>sc", ":nohls<CR>")
--}}}

--{{{ Goto Preview
map("n", "gpc", ':lua require("goto-preview").close_all_win()<CR>')
map("n", "gpd", ':lua require("goto-preview").goto_preview_definition()<CR>')
map("n", "gpi", ':lua require("goto-preview").goto_preview_implementation()<CR>')
--}}}

--{{{ Workspace management
map("n", "<leader>wa", vim.lsp.buf.add_workspace_folder, opts)
map("n", "<leader>wr", vim.lsp.buf.remove_workspace_folder, opts)
map("n", "<leader>wl", function()
	print(vim.inspect(vim.lsp.buf.list_workspace_folders()))
end, opts)
--}}}
